#include <stdlib.h>
#include <stdio.h>
#include <SDL2/SDL.h>

#include <time.h> // time() for srand().
#include <math.h> // sqrt().
#include <string.h>
#include <list>

#define CONSOLE_W 20
#define SCREEN_W 60
#define SCREEN_H 60
#define TILE_SIZE 10
#define SDL_BLENDMODE_DEFAULT SDL_BLENDMODE_MOD

/** TODO **\

* Shut down lights when the core generator is disabled.
* Display a line when something in sigh die.
* Explain how to regain life.
* Have a true inventory.

\** *  * **/

/** Enumerations **/

enum fsm_state {
	fsm_nothing,
	fsm_stand_and_avoid_light,
	fsm_blind_roam,
	fsm_search_for_kill,
	fsm_track_prey
};

/** Math and misc functions **/

bool isvalid(int x, int y) {
	return(0 <= y && y < SCREEN_H && 0 <= x && x < SCREEN_W);
}

inline int min(int a, int b) {
	return(a < b ? a : b);
}

inline int max(int a, int b) {
	return(a < b ? b : a);
}

inline float square(int x) {
	return(x*x);
}

inline float distance(int x1, int y1, int x2, int y2) {
	return(sqrt(square(x1-x2)+square(y1-y2)));
}

inline void* ec_malloc(int size) {
	void* ptr = malloc(size);
	if(ptr == NULL) {
		perror("allocation error");
		exit(EXIT_FAILURE);
	}
	return(ptr);
}

inline void putchar(SDL_Renderer* renderer, SDL_Texture * ascii[128], int x, int y, int c, int r, int g, int b) {
	static SDL_Rect dst = {0, 0, TILE_SIZE, TILE_SIZE};
	dst.x = x * TILE_SIZE;
	dst.y = y * TILE_SIZE;
	SDL_RenderCopy(renderer, ascii[c], NULL, &dst);
	SDL_SetRenderDrawColor(renderer, r, g, b, SDL_ALPHA_OPAQUE);
	SDL_RenderFillRect(renderer, &dst);
}

inline void ascii_to_dir(char c, int * xshift, int * yshift) {
	if(c <= 'Z') {
		c += ('z'-'Z');
	}
	switch(c) {
		case 'k': *xshift = 0;  *yshift = -1; break;
		case 'j': *xshift = 0;  *yshift = +1; break;
		case 'h': *xshift = -1; *yshift =  0; break;
		case 'l': *xshift = +1; *yshift =  0; break;
		case 'y': *xshift = -1; *yshift = -1; break;
		case 'u': *xshift = +1; *yshift = -1; break;
		case 'b': *xshift = -1; *yshift = +1; break;
		case 'n': *xshift = +1; *yshift = +1; break;
		default: *xshift = 0; *yshift = 0; break;
	}
}

/** SDL Functions **/

void init_sdl(SDL_Window ** window, SDL_Renderer ** renderer) {
	SDL_Init(SDL_INIT_VIDEO);

	// Create window and renderer.

	*window = SDL_CreateWindow(NULL, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 0, 0, 0);
	if(!*window) {
		fprintf(stderr, "Unable to open new window: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
	*renderer = SDL_CreateRenderer(*window, -1, 0);
	if(!*renderer) {
		fprintf(stderr, "Unable to create new renderer: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}

	// Configure window.

	SDL_SetWindowTitle(*window, "D3: ranDom Dark Death");
	// SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "linear");  // make the scaled rendering look smoother. // DON'T DO THAT!!!
	SDL_RenderSetLogicalSize(*renderer, (SCREEN_W+CONSOLE_W)*TILE_SIZE, SCREEN_H*TILE_SIZE);
	SDL_SetWindowSize(*window, (SCREEN_W+CONSOLE_W)*TILE_SIZE, SCREEN_H*TILE_SIZE);
	SDL_SetWindowPosition(*window, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
	SDL_ShowCursor(SDL_DISABLE);
#ifndef DEBUG
	SDL_SetWindowFullscreen(*window, SDL_WINDOW_FULLSCREEN);
#endif
	SDL_SetRenderDrawColor(*renderer, 0, 0, 0, 255);
	SDL_RenderClear(*renderer);
	SDL_RenderPresent(*renderer);
	SDL_SetRenderDrawBlendMode(*renderer, SDL_BLENDMODE_DEFAULT);

}

void quit_sdl(SDL_Window * window, SDL_Renderer * renderer, SDL_Texture * ascii[128]) {
	for(int i=0; i<128; i++) {
		SDL_DestroyTexture(ascii[i]);
	}
	SDL_DestroyRenderer(renderer);
	SDL_DestroyWindow(window);
	SDL_Quit();
}

void load_ascii(SDL_Renderer * renderer, SDL_Texture * ascii[128]) {
	SDL_Surface * ascii_complete ;
	SDL_Surface * surface;
	SDL_Rect src;

	// Load file.

	ascii_complete = SDL_LoadBMP("ascii.bmp");
	if(!ascii_complete) {
		fprintf(stderr, "Unable to load ascii bmp file: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}

	// Split ascii_complete into individual tiles.

	if(SDL_BYTEORDER == SDL_BIG_ENDIAN) {
		surface = SDL_CreateRGBSurface(0, TILE_SIZE, TILE_SIZE, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF);
	} else {
		surface = SDL_CreateRGBSurface(0, TILE_SIZE, TILE_SIZE, 32, 0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000);
	}
	if(!surface) {
		fprintf(stderr, "Unable to create Alpha surface: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
	src.w = src.h = TILE_SIZE;
	for(int i=0; i<128; i++) {
		src.y = (i/8)*TILE_SIZE;
		src.x = (i%8)*TILE_SIZE;
		SDL_BlitSurface(ascii_complete, &src, surface, NULL);
		ascii[i] = SDL_CreateTextureFromSurface(renderer, surface);
	}

	// Clean up.

	SDL_FreeSurface(ascii_complete);
	SDL_FreeSurface(surface);
}

/* Classes */

class level; // Forward declared.

class localized {
public:
	class level * level = NULL;
	int x = 0;
	int y = 0;
	localized() { }
	localized(class level * level) : level(level) {
		x = rand()%SCREEN_W;
		y = rand()%SCREEN_H;
// TODO : random spawn must take care of unpassable landing.
	}
	localized(class level * level, int x, int y) : level(level), x(x), y(y) { }
	/*
	void set(int X, int Y) {
		if(isvalid(X,Y)) {
			x = X;
			y = Y;
		}
	}
	*/
};

class low_drawable {
public:
	char aspect = ' ';
	int r = 255;
	int g = 255;
	int b = 255;
	int xshift = 0;
	low_drawable() { }
	low_drawable(char a) : aspect(a) { }
	low_drawable(char a, int r, int g, int b) : aspect(a), r(r), g(g), b(b) { }
	inline void draw(SDL_Renderer * renderer, SDL_Texture * ascii[128], int x, int y, float light) {
		if(light >= 1.0) {
			putchar(renderer, ascii, x+xshift, y, aspect, r, g, b);
		} else if(light > 0.0) {
			putchar(renderer, ascii, x+xshift, y, aspect, r*light, g*light, b*light);
		}
	}
};

class drawable : public low_drawable {
public:
	drawable() { xshift = CONSOLE_W; }
	drawable(char a) : low_drawable(a) { xshift = CONSOLE_W; }
	drawable(char a, int r, int g, int b) : low_drawable(a,r,g,b) { xshift = CONSOLE_W; }
};

class console {
public:
	class low_drawable board[SCREEN_H][CONSOLE_W];
	int topline = 0;
	// bool modified = true;
	SDL_Renderer * renderer = NULL;
	SDL_Texture ** ascii = NULL;
	console(SDL_Renderer * renderer, SDL_Texture ** ascii) : renderer(renderer), ascii(ascii) { }
	void draw() {
		// if(modified) {
			for(int i=topline; i<SCREEN_H; i++) {
				for(int j=0; j<CONSOLE_W; j++) {
					board[i][j].draw(renderer, ascii, j, i, (float) i/(SCREEN_H-topline));
				}
			}
			// modified = false;
		// }
	}
	void add_line(class low_drawable line[CONSOLE_W]) {
		int i;
		for(i=0; i<SCREEN_H-1; i++) {
			for(int j=0; j<CONSOLE_W; j++) {
				board[i][j] = board[i+1][j];
			}
		}
		for(int j=0; j<CONSOLE_W; j++) {
			board[i][j] = line[j];
		}
		// modified = true;
	}
	void add_str(const char* str, int r, int g, int b) {
		add_str(str, r, g, b, r, g, b);
	}
	void add_str(const char* str, int r, int g, int b, int r2, int g2, int b2) {
		class low_drawable line[CONSOLE_W];
		int i;
		for(i=0; i<CONSOLE_W && str[i] != '\0'; i++) {
			if(str[i] >= 'A' && str[i] <= 'Z') {
				line[i] = low_drawable(str[i], r2, g2, b2);
			} else {
				line[i] = low_drawable(str[i], r, g, b);
			}
		}
		for(; i<CONSOLE_W; i++) {
			line[i] = low_drawable();
		}
		add_line(line);
	}
	void add_str(const char* str) {
		class low_drawable line[CONSOLE_W];
		for(int i=0; i<CONSOLE_W && str[i] != '\0'; i++) {
			line[i] = low_drawable(str[i]);
		}
		add_line(line);
	}
	void begin_list(const char* header) {
		add_str("");
		add_str(header, 255, 0, 255);
	}
	void end_list() {
		add_str("");
		refresh();
	}
	void refresh() {
		draw();
		SDL_RenderPresent(renderer);
	}
	bool question_tag(const char* str) {
		SDL_Event event;
		bool answer;
		add_str("");
		add_str(str);
		add_str("[Y]es or [N]o", 255, 0, 255);
		refresh();
		while(1) {
			SDL_WaitEvent(&event);
			if(event.type == SDL_TEXTINPUT) {
				if(event.text.text[0] == 'y' || event.text.text[0] == 'Y') {
					answer = true;
					break;
				} else if(event.text.text[0] == 'n' || event.text.text[0] == 'N') {
					answer = false;
					break;
				} else {
					add_str("[Y]es or [N]o", 255, 0, 255);
					refresh();
				}
			}
		}
		add_str("");
		refresh();
		return(answer);
	}
	void press_any() {
		SDL_Event event;
		add_str("");
		add_str("Press any key.", 255, 0, 255);
		refresh();
		while(1) {
			SDL_WaitEvent(&event);
			if(event.type == SDL_TEXTINPUT || event.type == SDL_KEYDOWN) {
				break;
			}
		}
		add_str("");
		refresh();
	}
	void sound(const char* str) {
		add_str(str, 255, 0, 255);
	}
};

class named {
public:
	const char* name = "";
	named() { }
	named(const char* name) : name(name) { }
	void describe(class console * console, int r, int g, int b) {
		console->add_str(name, r, g, b);
	}
	void describe(class console * console) {
		console->add_str(name, 255, 255, 255);
	}
	void describe(class console * console, const char* prestr) {
		char buf[CONSOLE_W];
		strncpy(buf, prestr, CONSOLE_W);
		strncat(buf, name, CONSOLE_W-strlen(prestr));
		console->add_str(buf, 255, 255, 255);
	}
};

class tile : public drawable, public named {
public:
	bool passable = true;
	bool translucent = true;
	float light = 0.0; // 1.0 or more is fully bright. 0.0 is entirely dark.
	tile() : drawable('.', 255, 255, 255), named("floor") { }
	/*
	tile(char a, int r, int g, int b) : drawable(a,r,g,b), named("floor") { }
	tile(char a, int r, int g, int b, bool passable, bool translucent) :
		drawable(a,r,g,b), passable(passable), translucent(translucent)
	{
		if(!passable && !translucent) {
			name = "wall";
		} else if(!passable && translucent) {
			name = "glass";
		}
	}
	*/
	tile(char a, int r, int g, int b, const char* name) :
		drawable(a,r,g,b), named(name) { }
	tile(char a, int r, int g, int b, bool passable, bool translucent, const char* name) :
		drawable(a,r,g,b), named(name), passable(passable), translucent(translucent) { }
	inline void draw(SDL_Renderer * renderer, SDL_Texture * ascii[128], int x, int y) {
		drawable::draw(renderer, ascii, x, y, light);
	}
	void describe(class console * console) {
		named::describe(console, r, g, b);
	}
};

static class tile upstairs = tile('<', 128, 128, 128, "upward stairway");
static class tile downstairs = tile('>', 128, 128, 128, "downward stairway");
static class tile updownstairs = tile('X', 128, 128, 128, "up-down stairway");
static class tile tile_floor = tile('.', 255, 255, 255, "floor");
static class tile tile_wall = tile('#', 255, 255, 255, false, false, "wall");
static class tile tile_glass = tile('#', 0, 255, 255, false, true, "glass");

bool visible(int x1, int y1, int x2, int y2, class tile floor[SCREEN_W][SCREEN_H]) {
	for(int x=min(x1,x2)+1; x < max(x1,x2); x++) {
		int y = (y2-y1)*(x-x1)/(x2-x1)+y1;
		if(!floor[x][y].translucent) {
			return(false);
		}
	}
	for(int y = min(y1,y2)+1; y < max(y1,y2); y++) {
		int x = (x2-x1)*(y-y1)/(y2-y1)+x1;
		if(!floor[x][y].translucent) {
			return(false);
		}
	}
	return(true);
}

class light_source {
public:
	float intensity = 1.0; // 0.1 intensity is lost for each square light travel.
	float multiple = 1.0;
	light_source() { }
	light_source(float intensity) : intensity(intensity) { }
	void add_light_around(class tile floor[SCREEN_W][SCREEN_H], int x, int y) {
		int I = intensity*multiple;
		for(int i=max(x-I*10,0); i<min(x+I*10,SCREEN_W); i++) {
			for(int j=max(y-I*10,0); j<min(y+I*10,SCREEN_H); j++) {
				if(visible(x,y,i,j, floor)) {
					float light = I - distance(x,y,i,j) * 0.1;
					if(light > 0.0) {
						floor[i][j].light += light;
					}
				}
			}
		}
	}
};

class static_light : public light_source, public localized {
public:
	static_light(class level * l, int x, int y) : light_source(), localized(l,x,y) { }
	static_light(float intensity, class level * l, int x, int y) : light_source(intensity), localized(l,x,y) { }
	void add_light_around(); // Forward declared.
};

class equipment : public named {
public:
	unsigned int dammages = 0;
	equipment(const char* name) :
		named(name) { }
	equipment(const char* name, unsigned int dmg) :
		named(name), dammages(dmg) { }
	virtual ~equipment() { }
	virtual unsigned int get_dammages() {
		return(dammages);
	}
};

class weapon : public equipment {
public:
	weapon(const char* name) :
		equipment(name, 1) { }
	weapon(const char* name, unsigned int dmg) :
		equipment(name, dmg) { }
};

class weapon_grasp : public weapon {
public:
	class weapon * grasp = NULL;
	weapon_grasp() : weapon("main hand", 0) { }
	weapon_grasp(const char* name) : weapon(name, 0) { }
	weapon_grasp(const char* name, unsigned int dmg) : weapon(name, dmg) { }
	~weapon_grasp() {
		if(grasp != NULL) {
			delete(grasp);
		}
	}
	unsigned int get_dammages() {
		if(grasp == NULL) {
			return(dammages);
		} else {
			return(grasp->get_dammages());
		}
	}
	void describe(class console * console) {
		named::describe(console);
		if(grasp != NULL) {
			grasp->describe(console, "> ");
		}
	}
};

class main_hand : public weapon_grasp {
public:
	main_hand() : weapon_grasp() { }
};

class clawed_paw : public weapon_grasp {
public:
	clawed_paw() : weapon_grasp("clawed paw", 1) { }
};

class sword : public weapon {
public:
	sword() : weapon("sword") { }
};

class grafted_blade : public weapon {
public:
	grafted_blade() : weapon("grafted sword") { }
};

class light_equipment : public equipment, public light_source {
public:
	light_equipment(const char* name) :
		equipment(name) { }
	light_equipment(const char* name, float intensity) :
		equipment(name), light_source(intensity) { }
};

class bio_light : public light_equipment {
public:
	bio_light() : light_equipment("biolight", 0.5) { }
};

class gauge : public named {
public:
#define SCALE_NORMAL  -1
#define SCALE_REVERSE -2
	int max;
	int val;
	char aspect = '#';
	int r = SCALE_REVERSE;
	int g = SCALE_NORMAL;
	int b = 0;
	gauge(const char* name, int max) :
		named(name), max(max), val(max) { }
	gauge(const char* name, int max, char aspect) :
		named(name), max(max), val(max), aspect(aspect) { }
	gauge(const char* name, int max, char aspect, int r, int g, int b) :
		named(name), max(max), val(max), aspect(aspect), r(r), g(g), b(b) { }
	void increase(int i) {
		val += i;
		if(val > max) {
			val = max;
		} else if(val < 0) {
			val = 0;
		}
	}
	void decrease(int i) {
		increase(-i);
	}
	bool isempty() {
		return(val <= 0);
	}
	static int scale(int x, float n) {
		if(x == SCALE_REVERSE) {
			return(255*(1-n));
		} else if(x == SCALE_NORMAL) {
			return(255*n);
		} else {
			return(x);
		}
	}
	void draw(SDL_Renderer* rend, SDL_Texture * ascii[128], int y) {
		draw_name(rend, ascii, y);
		draw_val(rend, ascii, y+1);
		draw_gauge(rend, ascii, y+2);
	}
	void draw_name(SDL_Renderer* rend, SDL_Texture * ascii[128], int y) {
		static SDL_Rect dst = {0, 0, TILE_SIZE, TILE_SIZE};
		dst.y = y * TILE_SIZE;
		SDL_SetRenderDrawColor(rend,
				scale(r, (float)val/max),
				scale(g, (float)val/max),
				scale(b, (float)val/max),
				SDL_ALPHA_OPAQUE);
		for(int i=0; i<CONSOLE_W-2 && name[i] != '\0'; i++) {
			dst.x = (i+1) * TILE_SIZE;
			SDL_RenderCopy(rend, ascii[(int)name[i]], NULL, &dst);
			SDL_RenderFillRect(rend, &dst);
		}
	}
	void draw_val(SDL_Renderer* rend, SDL_Texture * ascii[128], int y) {
		int i=0;
		int j;
		char buf[32];
		sprintf(buf, "%d", val);
		for(j=0; i<CONSOLE_W-2 && buf[j] != '\0'; i++, j++) {
			putchar(rend, ascii, i+1, y, buf[j],
					scale(r, (float)val/max),
					scale(g, (float)val/max),
					scale(b, (float)val/max));
		}
		putchar(rend, ascii, i+1, y, '/', scale(r, 0.5), scale(g, 0.5), scale(b, 0.5));
		i++;
		sprintf(buf, "%d", max);
		for(j=0; i<CONSOLE_W-2 && buf[j] != '\0'; i++, j++) {
			putchar(rend, ascii, i+1, y, buf[j], scale(r, 1), scale(g, 1), scale(b, 1));
		}

	}
	void draw_gauge(SDL_Renderer* rend, SDL_Texture * ascii[128], int y) {
		int width = (val*(CONSOLE_W-2)/max);
		for(int i=0; i<width; i++) {
			putchar(rend, ascii, i+1, y, aspect,
					scale(r, (float)i/width),
					scale(g, (float)i/width),
					scale(b, (float)i/width));
		}
	}
};

class blood : public gauge {
public:
	blood(int max) : gauge("blood", max, '~', SCALE_REVERSE, SCALE_NORMAL, 0) { }
};

class ichor : public gauge {
public:
	ichor(int max) : gauge("ichor", max, '~', 255, 255, 255) { }
};

class steam_blood : public gauge {
public:
	steam_blood(int max) : gauge("vapour", max, '+', 255, SCALE_REVERSE, 0) { }
};

class ionic_blood : public gauge {
public:
	ionic_blood(int max) : gauge("ionic solution", max, '+', 0, SCALE_NORMAL, 255) { }
};

class sap : public gauge {
public:
	sap(int max) : gauge("sap", max, '~', SCALE_REVERSE, 255, SCALE_REVERSE) { }
};

class creature; // Forward declared.
class switchable; // Forward declared.
class floor_item; // Forward declared.
class consumable_item; // Forward declared.

class level {
public:
	class tile floor[SCREEN_W][SCREEN_H];
	std::list<class static_light *> lights;
	std::list<class creature *> creatures;
	std::list<class activable *> activables;
	std::list<class consumable_item *> items;
	class level * up = NULL;
	class level * down = NULL;
	float global_light = 0.0;
	level() { }
	level(class tile tile) {
		for(int x=0; x<SCREEN_W; x++) {
			for(int y=0; y<SCREEN_H; y++) {
				floor[x][y] = tile;
			}
		}
	}
	~level(); // Forward declared.
	class creature * get_creature_by_xy(int x, int y); // Forward declared.
	class consumable_item * get_item_by_xy(int x, int y); // Forward declared.
	void describe(int x, int y, class console * console); // Forward declared.
	void reset_light() {
		for(int i=0; i<SCREEN_W; i++) {
			for(int j=0; j<SCREEN_W; j++) {
				floor[i][j].light = global_light;
			}
		}
	}
	static void reset_screen(SDL_Renderer * renderer) {
		// XXX //
		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
		SDL_RenderClear(renderer);
		/* XXX //
		SDL_Rect dst;
		dst.x = CONSOLE_W*TILE_SIZE;
		dst.y = 0;
		dst.w = SCREEN_W*TILE_SIZE;
		dst.h = SCREEN_H*TILE_SIZE;
		SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_NONE);
		SDL_SetRenderDrawColor(renderer, 0, 0, 0, SDL_ALPHA_OPAQUE);
		SDL_RenderDrawRect(renderer, &dst);
		SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_DEFAULT);
		// XXX */
	}
	void link_up(int x, int y) {
		if(floor[x][y].aspect == '>') {
			floor[x][y] = updownstairs;
		} else {
			floor[x][y] = upstairs;
		}
		if(up != NULL) {
			if(up->floor[x][y].aspect == '<') {
				up->floor[x][y] = updownstairs;
			} else {
				up->floor[x][y] = downstairs;
			}
		}
	}
	void link_down(int x, int y) {
		if(floor[x][y].aspect == '<') {
			floor[x][y] = updownstairs;
		} else {
			floor[x][y] = downstairs;
		}
		if(down != NULL) {
			if(down->floor[x][y].aspect == '<') {
				down->floor[x][y] = updownstairs;
			} else {
				down->floor[x][y] = upstairs;
			}
		}
	}
	void add_random(class tile tile) {
		add_random(tile, 0.1);
	}
	void add_random(class tile tile, float probability) {
		for(int x=0; x<SCREEN_W; x++) {
			for(int y=0; y<SCREEN_H; y++) {
				if((float)rand()/RAND_MAX < probability) {
					floor[x][y] = tile;
				}
			}
		}
	}
	void add_circle(int xcenter, int ycenter, int radius, class tile tile) {
		add_random_in_circle(xcenter, ycenter, radius, tile, 1.0);
	}
	void add_random_in_circle(int xcenter, int ycenter, int radius, class tile tile, float probability) {
		for(int x=max(xcenter-radius,0); x<min(xcenter+radius,SCREEN_W); x++) {
			for(int y=max(ycenter-radius,0); y<min(ycenter+radius,SCREEN_H); y++) {
				if(distance(xcenter, ycenter, x, y) < radius && (float)rand()/RAND_MAX < probability) {
					floor[x][y] = tile;
				}
			}
		}
	}
	void add_path(int x1, int y1, int x2, int y2, class tile tile) {
		add_rand_in_path(x1, y1, x2, y2, tile, 1.0);
	}
	void add_rand_in_path(int x1, int y1, int x2, int y2, class tile tile, float probability) {
		for(int x=min(x1,x2)+1; x < max(x1,x2); x++) {
			int y = (y2-y1)*(x-x1)/(x2-x1)+y1;
			if((float)rand()/RAND_MAX < probability) {
				if(floor[x][y].aspect != '>' && floor[x][y].aspect != '<' && floor[x][y].aspect != 'X')
				floor[x][y] = tile;
			}
		}
		for(int y = min(y1,y2)+1; y < max(y1,y2); y++) {
			int x = (x2-x1)*(y-y1)/(y2-y1)+x1;
			if((float)rand()/RAND_MAX < probability) {
				if(floor[x][y].aspect != '>' && floor[x][y].aspect != '<' && floor[x][y].aspect != 'X')
				floor[x][y] = tile;
			}
		}
	}
	void add_splash(int xcenter, int ycenter, int radius, class tile tile) {
		for(int x=max(xcenter-radius,0); x<min(xcenter+radius,SCREEN_W); x++) {
			for(int y=max(ycenter-radius,0); y<min(ycenter+radius,SCREEN_H); y++) {
				if(distance(xcenter, ycenter, x, y) < rand()%radius) {
					floor[x][y] = tile;
				}
			}
		}
	}
	void add_splash_liquid(int xcenter, int ycenter, int radius, class tile tile) {
		for(int x=max(xcenter-radius,0); x<min(xcenter+radius,SCREEN_W); x++) {
			for(int y=max(ycenter-radius,0); y<min(ycenter+radius,SCREEN_H); y++) {
				if(distance(xcenter, ycenter, x, y) < rand()%radius &&
						visible(xcenter, ycenter, x, y, floor)) {
					if(!floor[x][y].passable ||
							floor[x][y].aspect == '<' ||
							floor[x][y].aspect == '>' ||
							floor[x][y].aspect == 'X') {
						floor[x][y].r = tile.r;
						floor[x][y].g = tile.g;
						floor[x][y].b = tile.b;
					} else {
						floor[x][y] = tile;
					}
				}
			}
		}
	}
	void add_irregular_rooms(int nb_rooms, class tile tile) {
		int x = rand()%SCREEN_W;
		int y = rand()%SCREEN_H;
		add_irregular_rooms(nb_rooms, tile, &x, &y);
	}
	void add_irregular_rooms(int nb_rooms, class tile tile, int* X, int* Y) {
		int matrix[SCREEN_W][SCREEN_H];
		int next[SCREEN_W][SCREEN_H];
		bool* graph = (bool*) ec_malloc(sizeof(bool)*nb_rooms*nb_rooms);
		int* center_x = (int*) ec_malloc(sizeof(int)*nb_rooms);
		int* center_y = (int*) ec_malloc(sizeof(int)*nb_rooms);
		int* vote = (int*) ec_malloc(sizeof(int)*nb_rooms);
		bool stop = false;

		// Init floor.
		for(int y=0; y<SCREEN_H; y++) {
			for(int x=0; x<SCREEN_W; x++) {
				matrix[x][y] = -1;
				next[x][y] = -1;
			}
		}

		// Plant room's seeds.
		matrix[*X][*Y] = 0;
		next[*X][*Y] = 0;
		for(int k=1; k<nb_rooms; k++) {
			int x = rand()%SCREEN_W;
			int y = rand()%SCREEN_H;
			matrix[x][y] = k;
			next[x][y] = k;
		}

		// Grow rooms.
		while(!stop) {
			stop = true;
			// For each room.
			for(int r=0; r<nb_rooms; r++) {
				for(int y=0; y<SCREEN_H; y++) {
					for(int x=0; x<SCREEN_W; x++) {
						if(matrix[x][y] == -1) {
							// Reset votes.
							for(int k=0; k<nb_rooms; k++) {
								vote[k] = 0;
							}
							// Count votes.
							for(int j=-1; j<= +1; j++) {
								for(int i=-1; i<= +1; i++) {
									// if((i==0 || j==0) && isvalid(x+i, y+j) && matrix[x+i][y+j] != -1) { }
									if(!(i==0 && j==0) && isvalid(x+i, y+j) && matrix[x+i][y+j] != -1) {
										(vote[matrix[x+i][y+j]])++;
									}
								}
							}

							int found = -1;
							int collision = 0;
							for(int k=0; k<nb_rooms; k++) {
								if(vote[k] > 0) {
									if(found == -1) {
										found = k;
									} else {
										collision = 1;
										break;
									}
								}
							}
							// If only one room is next, become this room.
							if(found == r && !collision) {
								next[x][y] = found;
								stop = 0;
							}
						}
					}
				}
				for(int y=0; y<SCREEN_H; y++) {
					for(int x=0; x<SCREEN_W; x++) {
						matrix[x][y] = next[x][y];
					}
				}
			} // End of "for each room".
		}
		// Add extern walls.
		for(int x=0; x<SCREEN_W; x++) {
			matrix[x][0] = matrix[x][SCREEN_H-1] = -1;
		}
		for(int y=0; y<SCREEN_H; y++) {
			matrix[0][y] = matrix[SCREEN_W-1][y] = -1;
		}
		// Calculate rooms' centers.
		for(int r=0; r<nb_rooms; r++) {
			int nb_tiles = 0;
			int xsum = 0;
			int ysum = 0;
			for(int y=0; y<SCREEN_H; y++) {
				for(int x=0; x<SCREEN_W; x++) {
					if(matrix[x][y] == r) {
						nb_tiles++;
						xsum += x;
						ysum += y;
					}
				}
			}
			if(nb_tiles > 0) {
				center_x[r] = xsum/nb_tiles;
				center_y[r] = ysum/nb_tiles;
			} else {
				center_x[r] = -1;
				center_y[r] = -1;
			}
		}
		// Place static lights on center of each room.
		for(int r=0; r<nb_rooms; r++) {
			if(center_x[r] != -1) {
				lights.push_back(new static_light(this, center_x[r], center_y[r]));
			}
		}
		// Calculate adjacent floors.
		for(int i=0; i<nb_rooms; i++) {
			for(int j=0; j<nb_rooms; j++) {
				graph[i*nb_rooms+j] = false;
			}
		}
		for(int y=0; y<SCREEN_H; y++) {
			for(int x=0; x<SCREEN_W; x++) {
				if(matrix[x][y] == -1) {
					int r1 = -1;
					int r2 = -1;
					for(int j=-1; j<= +1; j++) {
						for(int i=-1; i<= +1; i++) {
							if(isvalid(x+i, y+j) && matrix[x+i][y+j] != -1) {
								if(r1 == -1) {
									r1 = matrix[x+i][y+j];
								} else {
									r2 = matrix[x+i][y+j];
								}
							}
						}
					}
					if(r1 != -1 && r2 != -1) {
						graph[r1*nb_rooms+r2] = true;
						graph[r2*nb_rooms+r1] = true;
					}
				}
			}
		}
		// Open full doors.
		for(int r1 = 0; r1<nb_rooms; r1++) {
			for(int r2=r1+1; r2<nb_rooms; r2++) {
				if(graph[r1*nb_rooms+r2]) {
					for(int x=min(center_x[r1],center_x[r2])+1; x < max(center_x[r1],center_x[r2]); x++) {
						int y = (center_y[r2]-center_y[r1])*(x-center_x[r1])/(center_x[r2]-center_x[r1])+center_y[r1];
						matrix[x][y] = nb_rooms;
					}
					for(int y = min(center_y[r1],center_y[r2])+1; y < max(center_y[r1],center_y[r2]); y++) {
						int x = (center_x[r2]-center_x[r1])*(y-center_y[r1])/(center_y[r2]-center_y[r1])+center_x[r1];
						matrix[x][y] = nb_rooms;
					}
				}
			}
		}
		// Get "last" room.
		int R = (rand()%nb_rooms);
#ifdef DEBUG
		printf("Last room is %d/%d\n", R, nb_rooms);
#endif // DEBUG
		*X = center_x[R];
		*Y = center_y[R];
		// Place walls on floor.
		for(int y=0; y<SCREEN_H; y++) {
			for(int x=0; x<SCREEN_W; x++) {
				if(matrix[x][y] == -1) {
					floor[x][y] = tile;
				}
			}
		}
		// Clean up.
		free(graph);
		free(center_x);
		free(center_y);
		free(vote);
	}
	void add_square_rooms(int nb_rooms) {
	}
};

class graveyard : public level {
public:
	graveyard() : graveyard(0.03) { }
	graveyard(float probability) : level(tile('.', 128, 128, 0, "earth")) {
		global_light = 0.3;
		for(int x=0; x<SCREEN_W; x++) {
			for(int y=0; y<SCREEN_H; y++) {
				if((float)rand()/RAND_MAX < probability) {
					if(rand()%2) {
						floor[x][y] = tile('n', 128, 128, 128, false, false, "stone slab");
					} else {
						floor[x][y] = tile('+', 128, 128, 128, false, false, "stone cross");
					}
					if(isvalid(x, y+1)) {
						floor[x][y+1] = tile('#', 128, 128, 0, "grave");
					}
				}
			}
		}
	}
};

class forest : public level {
public:
	forest() : forest(0.2, 0.05) { }
	forest(float proba_tree, float proba_flower) : level(tile('.', 0, 128, 0, "grass")) {
		global_light = 0.3;
		add_random(tile('T', 0, 192, 0, false, false, "tree"), proba_tree/2);
		add_random(tile('T', 0, 128, 0, false, false, "pine"), proba_tree/2);
		add_random(tile('*', 255, 0, 0, "flower"), proba_flower/2);
		add_random(tile('*', 255, 255, 0, "flower"), proba_flower/2);
	}
};

class desert : public level {
public:
	desert() : desert(0.1, 0.04) { }
	desert(float proba_grass, float proba_cactus) : level(tile('.', 255, 255, 0, "sand")) {
		global_light = 0.3;
		add_random(tile('"', 0, 192, 0, "grass tuft"), proba_grass);
		add_random(tile('#', 0, 128, 0, false, false, "cactus"), proba_cactus);
	}
};

void static_light::add_light_around() {
	light_source::add_light_around(level->floor, x, y);
}

class switchable {
public:
	virtual ~switchable() { }
	virtual void set(bool state) = 0;
};

class switchable_reverse : public switchable {
public:
	class switchable * switchable;
	switchable_reverse(class switchable * s) : switchable(s) { }
	~switchable_reverse() {
		delete(switchable);
	}
	void set(bool state) {
		switchable->set(!state);
	}
};

class switchable_light : public switchable {
public:
	class light_source * light; // Don't delete it on destruction!!
	switchable_light(class light_source * light) : light(light) { }
	void set(bool state) {
		if(state) {
			light->multiple = 1.0;
		} else {
			light->multiple = 0.0;
		}
	}
};

class switchable_tile : public switchable, public localized {
public:
	class tile aspect_on  = tile('o', 0, 255, 0, "switch");
	class tile aspect_off = tile('o', 255, 0, 0, "switch");
	switchable_tile(class localized l) :
		switchable(), localized(l) { }
	switchable_tile(class localized l, class tile on, class tile off) :
		switchable(), localized(l), aspect_on(on), aspect_off(off) { }
    ~switchable_tile() { }
	void set(bool state) {
		if(state) {
			level->floor[x][y] = aspect_on;
		} else {
			level->floor[x][y] = aspect_off;
		}
	}
};

class switchable_sound : public switchable {
public:
	class console * console;
	const char* str_on;
	const char* str_off;
	switchable_sound(class console * console, const char* str) : switchable_sound(console, str, str) { }
	switchable_sound(class console * console, const char* on, const char* off) : console(console), str_on(on), str_off(off) { }
	void set(bool state) {
		if(state) {
			console->sound(str_on);
		} else {
			console->sound(str_off);
		}
	}
};

class switchable_victory : public switchable {
public:
	bool * victory;
	switchable_victory(bool * victory) : victory(victory) { }
	void set(bool state) {
		*victory = state;
	}
};

class activable : public localized {
public:
	bool state;
	std::list<class switchable *> switches;
	activable(class localized l, bool state) : localized(l), state(state) { }
	activable(class localized l) : activable(l, false) { }
	~activable() {
		for(class switchable * s : switches) {
			delete(s);
		}
	}
	void toogle() {
		if(state) {
			state = false;
		} else {
			state = true;
		}
		for(class switchable * s : switches) {
			s->set(state);
		}
	}
};

class lever : public switchable_tile, public activable {
public:
	lever(class localized l, class console * console) : lever(l, console, false) { }
	lever(class localized l, class console * console, bool state) :
		switchable_tile(l, tile('\\', 0, 255, 0, "lever on"), tile('/', 255, 0, 0, "lever off")),
		activable(l, state)
	{
		switches.push_back(this);
		switches.push_back(new switchable_sound(console, "*cling*", "*clang*"));
		this->set(state);
	}
};

class door : public switchable_tile {
public:
	door(class localized l) :
		switchable_tile(l, tile('-', 128, 128, 0, "door"), tile('+', 128, 128, 0, false, false, "door")) { }
};

/* XXX //
class door_two_button : public activable {
public:
	door_two_button(class localized l) : activable(l)
	{
		switches.push_back(new door(l)
	}
};
// XXX */

class energy_core_trigger : public activable {
public:
	class switchable_tile s_tile;
	class switchable_tile s_core;
	// class switchable_victory s_victory; // XXX
	class switchable_reverse s_victory;
	energy_core_trigger(bool * victory, class console * console, class localized l, class localized l_core) :
		activable(l, true),
		s_tile(l, tile('O', 0, 255, 0, "core trigger"), tile('O', 255, 0, 0, "core trigger")),
		s_core(l_core, tile('@', 0, 255, 255, false, false, "core generator"), tile('@', 128, 128, 128, false, false, "core generator")),
		s_victory(new switchable_victory(victory))
	{
		l_core.level->floor[l_core.x][l_core.y] = tile('@', 0, 255, 255, false, false, "core generator");
		switches.push_back(&s_tile);
		switches.push_back(&s_core);
		switches.push_back(&s_victory);
		switches.push_back(new switchable_sound(console, "ENERGY CORE ACTIVATE", "ENERGY CORE SHUTDOWN"));
		s_tile.set(true);
		s_tile.set(true);
	}
};

class core_entrance : public level {
public:
	core_entrance(int x, int y) : level(tile_wall) {
		add_path(x, y, SCREEN_W/2, SCREEN_H/2, tile_floor);
		link_down(SCREEN_W/2, SCREEN_H/2);
	}
};

class labcore : public level {
public:
	labcore(bool * victory, class console * console) : level(tile_wall) {
		int x=SCREEN_W/2;
		int y=SCREEN_H/2;
		add_circle(x, y, 10, tile_floor);
		add_random_in_circle(x, y, 10, tile('|', 255, 0, 0, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('|', 0, 255, 255, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('|', 0, 0, 255, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('-', 255, 255, 0, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('-', 0, 255, 0, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('-', 0, 0, 255, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('=', 255, 0, 0, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('=', 0, 255, 0, "wire"), 0.1);
		add_random_in_circle(x, y, 10, tile('=', 255, 0, 255, "wire"), 0.1);
		activables.push_back(new energy_core_trigger(victory, console, localized(this, x, y-10), localized(this, x, y-11)));
		link_up(x, y);
	}
};

struct prey_data {
	class creature * prey;
	int last_prey_x;
	int last_prey_y;
};

struct fsm_data {
	struct prey_data prey_data;
};

class creature : public drawable, public localized, public named {
public:
	class light_equipment * light = NULL;
	class weapon * weapon = NULL;
	// class weapon_grasp * weapon_grasp = NULL; // XXX
	class gauge life;
	struct fsm_data fsm_data;
	enum fsm_state state;
	bool dead = false;
	creature(class drawable aspect, class localized spawn, enum fsm_state fsm, class gauge life) :
		drawable(aspect), localized(spawn), named("monster"), life(life), state(fsm) { }
	creature(const char* name, class drawable aspect, class localized spawn, enum fsm_state fsm, class gauge life) :
		drawable(aspect), localized(spawn), named(name), life(life), state(fsm) { }
	virtual ~creature() {
		class tile fluid;
		fluid = tile(life.aspect, gauge::scale(life.r, 0.0), gauge::scale(life.g, 0.0), gauge::scale(life.b, 0.0), life.name);
		if(level != NULL) {
			level->add_splash_liquid(x, y, 3, fluid);
		}
		if(light != NULL) {
			delete(light);
		}
		if(weapon != NULL) {
			delete(weapon);
		}
	}
	inline void draw(SDL_Renderer * renderer, SDL_Texture * ascii[128], float light) {
		drawable::draw(renderer, ascii, x, y, light);
	}
	void describe(class console * console) {
		named::describe(console, r, g, b);
		display_inventory(console);
		console->add_str("");
	}
	unsigned int get_dammages() {
		unsigned int dmg = 0;
		if(weapon != NULL) {
			dmg += weapon->get_dammages();
		}
		return(dmg);
	}
	void draw_all_gauges(SDL_Renderer * renderer, SDL_Texture * ascii[128]) {
		life.draw(renderer, ascii, 2);
	}
	int get_gauge_yshift() {
		const int nb_gauges = 1;
		const int gauge_size = 3;
		return(1+nb_gauges*(gauge_size+1)+1);
	}
	void display_inventory(class console * console) {
		if(weapon != NULL) {
			weapon->describe(console);
		}
		if(light != NULL) {
			light->describe(console);
		}
	}
	void hit(unsigned int dammages) {
		life.decrease(dammages);
		if(life.isempty()) {
			die();
		}
	}
	virtual void die() {
		dead = true;
	}
	void turn() {
		switch(state) {
			case fsm_nothing: break;
			case fsm_blind_roam: move_random(); break;
			case fsm_stand_and_avoid_light:
				if(level->floor[x][y].light < 0.3) {
					move_random();
				} else {
					move_into_darkness();
				}
				break;
			case fsm_search_for_kill:
				{
					class creature * prey = nereast_seen_creature();
					if(prey == NULL) {
						if(fsm_data.prey_data.prey == NULL) { // Search prey.
							move_random();
						} else { // Prey lost. Start search.
							state = fsm_track_prey;
							turn();
						}
					} else { // Prey targeted. Run after it.
						fsm_data.prey_data = {prey, prey->x, prey->y};
						move_toward_stupid(prey);
					}
				}
				break;
			case fsm_track_prey:
				if(fsm_data.prey_data.last_prey_x == x && fsm_data.prey_data.last_prey_y == y) {
					// End of track. Return to search.
					fsm_data.prey_data = {NULL, 0, 0};
					state = fsm_search_for_kill;
					turn();
				} else { // Follow track.
					move_toward_stupid(fsm_data.prey_data.last_prey_x, fsm_data.prey_data.last_prey_x);
				}
				break;
		}
	}
	void move(int xshift, int yshift) {
		if(!(xshift == 0 && yshift == 0)) {
			int nextx = x + xshift;
			int nexty = y + yshift;
			class creature * target = level->get_creature_by_xy(nextx, nexty);
			if(target != NULL) {
				target->hit(get_dammages());
			} else if(isvalid(nextx, nexty) && level->floor[nextx][nexty].passable) {
				y = nexty;
				x = nextx;
			}
		}
	}
	void move_up() {
		if((level->floor[x][y].aspect == '<' || level->floor[x][y].aspect == 'X') && level->up != NULL) {
			level->creatures.remove(this);
			level = level->up;
			level->creatures.push_back(this);
		}
	}
	void move_down() {
		if((level->floor[x][y].aspect == '>' || level->floor[x][y].aspect == 'X') && level->down != NULL) {
			level->creatures.remove(this);
			level = level->down;
			level->creatures.push_back(this);
		}
	}
	void move_into_darkness() {
		if(level->floor[x][y].light >= 0.3) {
			float less_light = level->floor[x][y].light;
			int xshift = 0;
			int yshift = 0;
			for(int i=-1; i<=+1; i++) {
				for(int j=-1; j<=+1; j++) {
					if(isvalid(x+i,y+j) && level->floor[x+i][y+j].passable &&
							level->floor[x+i][y+j].light < less_light) {
						xshift = i;
						yshift = j;
					}
				}
			}
			move(xshift, yshift);
		}
	}
	void move_random() {
		int xshift;
		int yshift;
		int givup = 0;
		do {
			if(++givup > 6) break;
			xshift = (rand()%3)-1;
			yshift = (rand()%3)-1;
		} while(!(isvalid(x+xshift,y+yshift) && level->floor[x+xshift][y+yshift].passable));
		move(xshift, yshift);
	}
	class creature * nereast_seen_creature() {
		float min_dist = SCREEN_W+SCREEN_H; // Sure it's enough.
		class creature * prey = NULL;
		for(class creature * target : level->creatures) {
			if(target != this &&
					distance(x, y, target->x, target->y) < min_dist &&
					visible(x, y, target->x, target->y, level->floor)) {
				prey = target;
			}
		}
		return(prey);
	}
	void move_toward_stupid(int tx, int ty) {
		int xshift = 0;
		int yshift = 0;
		int nextx, nexty;
		if(tx < x) {
			xshift = -1;
		} else if(tx > x) {
			xshift = +1;
		}
		if(ty < y) {
			yshift = -1;
		} else if(ty > y) {
			yshift = +1;
		}
		nextx = x+xshift;
		nexty = y+yshift;
		if(isvalid(nextx, nexty) && level->floor[nextx][nexty].passable) {
			move(xshift, yshift);
		} else {
// TODO : avoid obstacles.
		}
	}
	void move_toward_stupid(class creature * target) {
		move_toward_stupid(target->x, target->y);
	}
};

class consumable {
public:
	// virtual bool consume(class creature * consumer) = 0; // XXX
	virtual bool consume(creature * consumer, class console * console) = 0;
};

class steam_creature : public creature {
public:
	char fullname[CONSOLE_W];
	steam_creature(const char* name, char aspect, class localized l, enum fsm_state fsm, int life_max) :
		creature(fullname, drawable(aspect, 255, 128, 0), l, fsm, steam_blood(life_max))
	{
		strncpy(fullname, "steam-", CONSOLE_W);
		strncat(fullname, name, strlen(fullname)-CONSOLE_W);
	}
	virtual void die(); // Forward declared.
};

class ionic_creature : public creature {
public:
	char fullname[20];
	ionic_creature(const char* name, char aspect, class localized l, enum fsm_state fsm, int life_max) :
		creature(fullname, drawable(aspect, 0, 128, 255), l, fsm, ionic_blood(life_max))
	{
		strncpy(fullname, "ionic-", CONSOLE_W);
		strncat(fullname, name, strlen(fullname)-CONSOLE_W);
	}
	virtual void die(); // Forward declared.
};

class sap_creature : public creature {
public:
	char fullname[20];
	sap_creature(const char* name, char aspect, class localized l, enum fsm_state fsm, int life_max) :
		creature(fullname, drawable(aspect, 128, 255, 128), l, fsm, sap(life_max))
	{
		strncpy(fullname, "plant ", CONSOLE_W);
		strncat(fullname, name, strlen(fullname)-CONSOLE_W);
	}
	virtual void die(); // Forward declared.
};

class bug_creature : public creature {
public:
	bug_creature(const char* name, char aspect, class localized l, enum fsm_state fsm, int life_max) :
		creature(name, drawable(aspect, 128, 64, 0), l, fsm, ichor(life_max)) { }
};

class blood_creature : public creature {
public:
	blood_creature(const char* name, char aspect, class localized l, enum fsm_state fsm, int life_max) :
		creature(name, drawable(aspect, 255, 255, 255), l, fsm, blood(life_max)) { }
	virtual void die(); // Forward declared.
};

class rat : public blood_creature {
public:
	rat(class localized l) : blood_creature("large rat", 'r', l, fsm_stand_and_avoid_light, 4) {
		r = g = 128;
		b = 0;
	}
};

class human : public blood_creature {
public:
	human(class localized l) : blood_creature("human", '@', l, fsm_search_for_kill, 6) {
		class weapon_grasp * hand = new main_hand();
		hand->grasp = new sword();
		weapon = hand;
	}
};

class steam_bot : public steam_creature {
public:
	steam_bot(class localized l) : steam_creature("bot", '@', l, fsm_search_for_kill, 6) {
		class weapon_grasp * hand = new main_hand();
		hand->grasp = new sword();
		weapon = hand;
	}
};

class ionic_bot : public ionic_creature {
public:
	ionic_bot(class localized l) : ionic_creature("bot", '@', l, fsm_search_for_kill, 6) {
		class weapon_grasp * hand = new main_hand();
		hand->grasp = new sword();
		weapon = hand;
	}
};

class hybrid : public blood_creature {
public:
	hybrid(class localized l) : blood_creature("hybrid", '@', l, fsm_search_for_kill, 6) {
		weapon = new grafted_blade();
		r = 255;
		g = b = 128;
	}
};

class plant_man : public sap_creature {
public:
	plant_man(class localized l) : sap_creature("man", '@', l, fsm_search_for_kill, 6) {
		class weapon_grasp * hand = new main_hand();
		hand->grasp = new sword();
		weapon = hand;
	}
};

class ant_man : public bug_creature {
public:
	ant_man(class localized l) : bug_creature("ant-man", '@', l, fsm_search_for_kill, 6) {
		class weapon_grasp * hand = new main_hand();
		hand->grasp = new sword();
		weapon = hand;
	}
};

class beetle : public bug_creature {
public:
	beetle(class localized l) : bug_creature("beetle", '*', l, fsm_blind_roam, 2) { }
};

class glowing_beetle : public beetle {
public:
	glowing_beetle(class localized l) : beetle(l) {
		name = "glowing beetle";
		r = g = 255;
		b = 0;
		light = new bio_light();
	}
};

class ionic_spider : public ionic_creature {
public:
	ionic_spider(class localized l) : ionic_creature("spider", '*', l, fsm_blind_roam, 2) { }
};

class steam_spider : public steam_creature {
public:
	steam_spider(class localized l) : steam_creature("spider", '*', l, fsm_blind_roam, 2) { }
};

class creature * generate_random_player(class localized l) {
	class creature * player;
	switch(rand()%5) {
		case 0: player = new steam_bot(l); break;
		case 1: player = new ionic_bot(l); break;
		case 2: player = new plant_man(l); break;
		case 3: player = new ant_man(l); break;
		case 4:
		default:player = new human(l);
	}
	player->state = fsm_nothing;
	player->light = new light_equipment("lanturn", 1.0);
	return(player);
}

class floor_item : public localized, public drawable, public named {
public:
	floor_item(class localized l, char aspect, int r, int g, int b, const char* name) :
		localized(l),
		drawable(aspect, r, g, b),
		named(name) { }
	inline void draw(SDL_Renderer * renderer, SDL_Texture * ascii[128], float light) {
		drawable::draw(renderer, ascii, x, y, light);
	}
	void describe(class console * console) {
		named::describe(console, r, g, b);
	}
};

class consumable_item : public floor_item, public consumable {
public:
	consumable_item(class floor_item item) :
		floor_item(item) { }
	virtual ~consumable_item() { }
};

class food : public consumable_item {
public:
	int value;
	food(int value, class floor_item item) :
		consumable_item(item),
		value(value) { }
	bool consume(creature * consumer, class console * console) {
		if(consumer->life.aspect == '~') {
			consumer->life.increase(value);
			console->add_str("you eat");
			describe(console);
			return(true);
		} else {
			return(false);
		}
	}
};

class steam_supply : public consumable_item {
public:
	int value;
	steam_supply(int value, class floor_item item) :
		consumable_item(item),
		value(value) { }
	bool consume(creature * consumer, class console * console) {
		if(consumer->life.aspect == '+' && consumer->life.r == 255) {
			consumer->life.increase(value);
			console->add_str("you refuel with");
			describe(console);
			return(true);
		} else {
			return(false);
		}
	}
};

class ionic_supply : public consumable_item {
public:
	int value;
	ionic_supply(int value, class floor_item item) :
		consumable_item(item),
		value(value) { }
	bool consume(creature * consumer, class console * console) {
		if(consumer->life.aspect == '+' && consumer->life.b == 255) {
			consumer->life.increase(value);
			console->add_str("you refuel with");
			describe(console);
			return(true);
		} else {
			return(false);
		}
	}
};

class raw_meat : public food {
public:
	raw_meat(class localized l, int value) :
		food(value, floor_item(l, '%', 255, 0, 0, "raw meat")) { }
};

class vegetable : public food {
public:
	vegetable(class localized l, int value) :
		food(value, floor_item(l, '%', 0, 255, 0, "vegetable")) { }
};

class ionic_cell : public ionic_supply {
public:
	ionic_cell(class localized l, int value) :
		ionic_supply(value, floor_item(l, '%', 0, 0, 255, "ionic cell")) { }
};

class charcoal : public ionic_supply {
public:
	charcoal(class localized l, int value) :
		ionic_supply(value, floor_item(l, '%', 128, 128, 128, "charcoal")) { }
};

class laboratory {
public:
	unsigned int nb_levels = 0;
	class level * top = NULL;
	class level * bottom = NULL;
	laboratory() { }
	~laboratory() {
		if(top != NULL) {
			delete(top);
		}
	}
	void add_level(class level * level) {
		level->down = NULL;
		if(bottom == NULL) {
			top = bottom = level;
			level->up = NULL;
		} else {
			bottom->down = level;
			level->up = bottom;
			bottom = level;
		}
	}
};

// Forwarded.

void blood_creature::die() {
	level->items.push_back(new raw_meat(localized(level, x, y), life.max/2));
	creature::die();
}

void steam_creature::die() {
	level->items.push_back(new charcoal(localized(level, x, y), life.max/2));
	creature::die();
}

void ionic_creature::die() {
	level->items.push_back(new ionic_cell(localized(level, x, y), life.max/2));
	creature::die();
}

void sap_creature::die() {
	level->items.push_back(new vegetable(localized(level, x, y), life.max/2));
	creature::die();
}

/* XXX //
void ionic_creature::die() {
	level->items.push_back(new ionic_cell(localized(level, x, y), life.max/2));
	creature::die();
}
// XXX */

class creature * level::get_creature_by_xy(int x, int y) {
	for(class creature * ptr : creatures) {
		if(ptr->x == x && ptr->y == y) {
			return(ptr);
		}
	}
	return(NULL);
}

class consumable_item * level::get_item_by_xy(int x, int y) {
	for(class consumable_item * ptr : items) {
		if(ptr->x == x && ptr->y == y) {
			return(ptr);
		}
	}
	return(NULL);
}

level::~level() {
	if(down != NULL) {
		delete(down);
	}
	for(class creature * ptr : creatures) {
		delete(ptr);
	}
	for(class static_light * ptr : lights) {
		delete(ptr);
	}
	for(class activable * ptr : activables) {
		delete(ptr);
	}
	for(class consumable_item * ptr : items) {
		delete(ptr);
	}
}

void level::describe(int x, int y, class console * console) {
	if(isvalid(x, y)) {
		class creature * creature = get_creature_by_xy(x, y);
		if(creature == NULL) {
			class consumable_item * item = get_item_by_xy(x, y);
			if(item == NULL) {
				floor[x][y].describe(console);
			} else {
				item->describe(console);
			}
		} else {
			creature->describe(console);
		}
	}
}

/** Main **/

int main (int argc, char** argv) {
	SDL_Renderer * renderer;
	SDL_Window * window;
	SDL_Texture * ascii[128];
	SDL_Event event;
	class laboratory labo;
	class creature * player;
	class console my_console(NULL, NULL);
	int x;
	int y;
	bool stop = false;
	bool end_turn = false;
	bool victory = false;

	init_sdl(&window, &renderer);
	load_ascii(renderer, ascii);
	my_console = console(renderer, ascii);
	srand(time(NULL));

	/** Display Intro **/

	const char* intro[] = {
	"ranDom Dark Death   ", // 0
	"          By Pumpkin", // 1
	"for the             ",
	"Procedural Death Jam", // 3
	"          march 2014",
	"                    ",
	"You are  a member of",
	"     the LODGE      ", // 7
	"a  secret gathering,",
	"eternal enemy of the",
	"  DAEMONIS MACHINA  ", // 10
	"an  evil council  of",
	"corrupted alchemists",
	"and techno-mages.   ",
	"                    ",
	"the LODGE discovered", // 15
	"an old laboratory of",
	"the DAEMONIS MACHINA", // 17
	"    and sent YOU    ", // 18
	"to purge this place,",
	"deactivate  the main",
	"energy generator and",
	"seal the lab's entry",
	"by escaping it.     ",
	"                    ",
	"The entrance must be",
	"nearby. Search a '>'",
	"and press > upon it.",
	"Follow  the path of ",
	"white slabs.        ",
	""};

	for(int i=0; intro[i][0] != '\0'; i++) {
		if(i == 0) { // ranDom Dark Death
			my_console.add_str(intro[i], 192, 0, 192, 255, 32, 255);
		} else if(i == 1) { // By Pumpkin
			my_console.add_str(intro[i], 255, 128, 0);
		} else if(i == 3) { // Procedural Death Jam
			my_console.add_str(intro[i], 0, 0, 255, 0, 128, 255);
		} else if(i == 7 || i == 15) { // LODGE
			my_console.add_str(intro[i], 255, 255, 255, 0, 255, 0);
		} else if(i == 10 || i == 17) { // DAEMONIS MACHINA
			my_console.add_str(intro[i], 255, 255, 255, 192, 0, 0);
		} else if(i == 18) { // YOU
			my_console.add_str(intro[i], 255, 255, 255, 0, 255, 255);
		} else {
			my_console.add_str(intro[i]);
		}
	}

	/** Init laboratory **/

	int total_depth = 3+rand()%3;
	x = rand()%(SCREEN_W-10)+5;
	y = rand()%(SCREEN_H-10)+5;
	switch(rand()%3) {
		case 0: labo.add_level(new desert); break;
		case 1: labo.add_level(new forest); break;
		case 2:
		default:labo.add_level(new graveyard); break;
	}
	labo.top->add_rand_in_path(SCREEN_W/2, SCREEN_H/2, x, y,
                            // tile('+', 128, 128, 128, "slab"), 0.2);
                            tile('.', 255, 255, 255, "slab"), 0.2);
	for(int depth=0; depth<total_depth; depth++) {
#ifdef DEBUG
        printf("at depth %d/%d : x = %d ; y = %d\n", depth, total_depth, x, y);
#endif // DEBUG
		labo.add_level(new level);
		labo.bottom->link_up(x, y);
		labo.bottom->add_irregular_rooms(10, tile_wall, &x, &y);
		for(int i=0; i<10; i++) { /** Add 10 of each vermin at each level **/
			labo.bottom->creatures.push_back(new rat(localized(labo.bottom)));
			labo.bottom->creatures.push_back(new ionic_spider(localized(labo.bottom)));
			labo.bottom->creatures.push_back(new steam_spider(localized(labo.bottom)));
		}
		for(int j=0; j<3; j++) { /** Add 3 humanoid at each level **/
			switch(rand()%3) {
				case 0: labo.bottom->creatures.push_back(new hybrid(localized(labo.bottom))); break;
				case 1: labo.bottom->creatures.push_back(new ionic_bot(localized(labo.bottom))); break;
				case 2:
				default:labo.bottom->creatures.push_back(new steam_bot(localized(labo.bottom))); break;
			}
		}
	}
#ifdef DEBUG
        printf("at bottom : x = %d ; y = %d\n", x, y);
#endif // DEBUG
	labo.add_level(new core_entrance(x, y));
	labo.bottom->link_up(x, y);
	labo.add_level(new labcore(&victory, &my_console));

	/* XXX //
	labo.add_level(new level());
	labo.bottom->add_irregular_rooms(10, tile_wall);
	for(class static_light * light : labo.bottom->lights) {
		class lever * l = new lever(localized(labo.bottom, light->x, light->y), &my_console, true);
		l->switches.push_back(new switchable_light(light));
		labo.bottom->activables.push_back(l);
	}
	for(int i=0; i<3; i++) {
		int x = rand()%SCREEN_W;
		int y = rand()%SCREEN_H;
		labo.bottom->activables.push_back(new lever(localized(labo.bottom, x, y), &my_console));
		// labo.bottom->switches.push_back(new switchable_all_level_light(localized(labo.bottom, x, y)));
	}

	labo.add_level(new level());
	labo.bottom->add_splash(20, 20, 5, tile('~', 255, 0, 0, "blood"));
	labo.bottom->add_random(tile('#', 128, 0, 128, false, false, "pillar"));
	labo.bottom->add_random(tile('_', 0, 192, 0, "slab"));
	labo.bottom->link_up(20, 20);
	labo.bottom->lights.push_back(new static_light(labo.bottom, 20, 20));
	for(int i=0; i<10; i++) {
		labo.bottom->creatures.push_back(new beetle(localized(labo.bottom)));
	}
	for(int i=0; i<10; i++) {
		labo.bottom->creatures.push_back(new rat(localized(labo.bottom)));
	}
	labo.bottom->creatures.push_back(new steam_bot(localized(labo.bottom)));
	labo.bottom->creatures.push_back(new ionic_bot(localized(labo.bottom)));
	labo.bottom->creatures.push_back(new glowing_beetle(localized(labo.bottom)));

	labo.add_level(new level(tile('#', 128, 128, 128, false, false, "rock wall")));
	labo.bottom->add_splash(20, 20, 20, tile('.', 128, 128, 64, "earth"));
	labo.bottom->add_splash(20, 20, 20, tile('.', 128, 128, 64, "earth"));
	labo.bottom->add_circle(20, 20, 5, tile('.', 64, 255, 0, "grass"));
	labo.bottom->link_up(20, 20);
	// XXX */

	player = generate_random_player(localized(labo.top, SCREEN_W/2, SCREEN_H/2));
	labo.top->creatures.push_back(player);

	my_console.begin_list("You are a");
	player->describe(&my_console);
	player->life.max += 4; // XXX
	player->life.val += 4; // XXX
	my_console.end_list();
	my_console.add_str("Press '?' for help. ");

	/** Loop **/

	while(!stop) {

		/** Recalculate light **/

		player->level->reset_light();
		for(class static_light * light : player->level->lights) {
			light->add_light_around();
		}
		for(class creature * creature : player->level->creatures) {
			if(creature->light != NULL) {
				creature->light->add_light_around(creature->level->floor, creature->x, creature->y);
			}
		}

		/** Draw screen **/

		level::reset_screen(renderer);

		for(int x=0; x<SCREEN_W; x++) {
			for(int y=0; y<SCREEN_H; y++) {
				if(player->level->floor[x][y].light > 0.0 &&
						visible(player->x, player->y, x, y, player->level->floor)) {
					player->level->floor[x][y].draw(renderer, ascii, x, y);
				}
			}
		}

		for(class consumable_item * item : player->level->items) {
			if(visible(player->x, player->y, item->x, item->y, item->level->floor)) {
				item->draw(renderer, ascii, player->level->floor[item->x][item->y].light);
			}
		}

		for(class creature * creature : player->level->creatures) {
			if(visible(player->x, player->y, creature->x, creature->y, player->level->floor)) {
				creature->draw(renderer, ascii, player->level->floor[creature->x][creature->y].light);
			}
		}

		/** Draw GUI **/

		player->draw_all_gauges(renderer, ascii);
		my_console.topline = player->get_gauge_yshift() + 1;
		my_console.draw();

		/** Present **/

		SDL_RenderPresent(renderer);

		/** Process player input **/

		end_turn = false;
		while(!end_turn) {
			SDL_WaitEvent(&event);
			end_turn = true;
			if(event.type == SDL_QUIT || (event.type == SDL_KEYDOWN && event.key.keysym.sym == SDLK_ESCAPE)) {
				stop = true;
			} else if(event.type == SDL_TEXTINPUT) {
				switch(event.text.text[0]) {
					case 'q':
						if(my_console.question_tag("Really exit?")) {
							stop = true;
						}
						break;
					case 'Q':
						stop = true;
						break;
					case 'k':
					case 'j':
					case 'h':
					case 'l':
					case 'y':
					case 'u':
					case 'b':
					case 'n':
						{
							int x, y;
							ascii_to_dir(event.text.text[0], &x, &y);
							player->move(x, y);
						}
						break;
					case 'K':
					case 'J':
					case 'H':
					case 'L':
					case 'Y':
					case 'U':
					case 'B':
					case 'N':
						{
							int x, y;
							ascii_to_dir(event.text.text[0], &x, &y);
							player->level->describe(player->x+x, player->y+y, &my_console);
						}
						my_console.refresh();
						end_turn = false;
						break;
					case 't':
						if(player->light != NULL) {
							if(player->light->multiple <= 0.0) {
								player->light->multiple = 1.0;
							} else {
								player->light->multiple = 0.0;
							}
						}
						break;
					case 'a':
						for(class activable * s : player->level->activables) {
							if(s->x == player->x && s->y == player->y/*&& s->level == player->level*/) {
								s->toogle();
								// break;
							}
						}
						break;
					case ',':
						// TODO : loot.
						{
							class consumable_item * item = player->level->get_item_by_xy(player->x, player->y);
							if(item != NULL) {
								if(item->consume(player, &my_console)) {
									player->level->items.remove(item);
								}
							}
						}
						break;
					case 's': /** Sleep (aka wait) **/
						break;
					case 'i': /** Inventory **/
						my_console.begin_list("[Your inventory:]");
						player->display_inventory(&my_console);
						my_console.end_list();
						end_turn = false;
						break;
					case '?':
						my_console.begin_list("[HELP]");
						my_console.add_str("hjklyubn: move");
						my_console.add_str("HJKLYUBN: examine");
						my_console.add_str(">: move down");
						my_console.add_str("<: move up");
						my_console.add_str("t: toogle light");
						my_console.add_str("a: activate");
						my_console.add_str(",: loot");
						my_console.add_str("s: wait 1 turn");
						my_console.add_str("i: display inventory");
						my_console.add_str("?: display this help");
						my_console.add_str("q: quit");
						my_console.add_str("Q: force quit");
						my_console.end_list();
						end_turn = false;
						break;
					case '>':
						player->move_down();
						break;
					case '<':
						player->move_up();
						if(player->level == labo.top && victory == true) {
							stop = true;
						}
						break;
					default:
						end_turn = false;
				}
			} else {
				end_turn = false;
			}
		}

		/** Creatures act **/

		for(auto it = player->level->creatures.begin(); it != player->level->creatures.end(); it++) {
			if(!(*it)->dead) {
				(*it)->turn();
			}
		}

		/** Remove dead creatures **/

		for(auto it = player->level->creatures.begin(); it != player->level->creatures.end(); it++) {
			if((*it)->dead) {
				if(*it == player) {
					stop = true;
				} else {
					delete(*it);
					it = player->level->creatures.erase(it);
					continue;
				}
			}
		}

		// if(stop) { break; }
	}

	/** Outro **/

	if(victory) {
		const char* outro[] = {
			"  YOU found your way", // 0
			"in this  dark place,",
			"managed to avoid its",
			"dangers and its evil",
			"creatures, shut down",
			"its  power  reactor,",
			"sealed  its entrance",
			"and  abandoned it to",
			"its ruin.           ",
			"                    ",
			"  The LODGE is proud", // 10
			"of you. The DAEMONIS", // 11
			"MACHINA has lost one", // 12
			"of its labs, but how",
			"many are left? Would",
			"your quest even end?",
			""};
		my_console.add_str("");
		my_console.add_str("       VICTORY      ", 0, 255, 0);
		my_console.add_str("");
		for(int i=0; outro[i][0] != '\0'; i++) {
			if(i == 10) { // LODGE
				my_console.add_str(outro[i], 255, 255, 255, 0, 255, 0);
			} else if(i == 11 || i == 12) { // DAEMONIS MACHINA
				my_console.add_str(outro[i], 255, 255, 255, 192, 0, 0);
			} else if(i == 0) { // YOU
				my_console.add_str(outro[i], 255, 255, 255, 0, 255, 255);
			} else {
				my_console.add_str(outro[i]);
			}
		}
		my_console.press_any();
	} else {
		my_console.add_str("You die...", 255, 0, 0);
		my_console.press_any();
	}

	/** Clean up **/

	quit_sdl(window, renderer, ascii);
	return(EXIT_SUCCESS);
}

